from mock import Mock

from datarunner import Workflow, Flow, Step


def test_step():
    class Count(Step):
        def run(self):
            return [1, 2, 3]

    count = Count()
    assert count() == [1, 2, 3]       # Callable
    assert count.run() == [1, 2, 3]   # Runnable
    assert list(count) == [1, 2, 3]   # Iterable

    assert Step(lambda: 1).name == '<lambda>'
    assert Step(print).name == 'print'
    assert Step(Mock(name='mocked')).name == 'mocked'

    class Step1(Step):
        pass
    assert Step.instance(Step1()).name == 'Step1'

    class Step2(Step):
        def __str__(self):
            return 'Super Step'
    assert Step.instance(Step2()).name == 'Super Step'

    class Step3(Step):
        name = "Fast Step"
    assert Step.instance(Step3()).name == 'Fast Step'


def test_flow(capsys):
    flow = Flow(range(3), print)
    assert len(flow) == 2

    flow >> range(6) >> print
    assert len(flow) == 4

    flow()
    out, err = capsys.readouterr()
    print(out)
    assert """\
range(0, 3)
>> print
range(0, 3)
>> range(0, 6)
>> print
range(0, 6)
""" == out


def test_workflow_basics(capsys):
    class Load(Step):
        def __repr__(self):
            return 'Load("dest")'

        def run(self, num):
            return num * 2

    class Finish(Step):
        def run(self):
            print('finished')

    transform = Mock(name='transform', return_value=3)
    result = Workflow(lambda: print("started"),
                      [lambda: 1, transform, Load()],
                      etl=[lambda: 2, transform, Load()],
                      ).run()

    transform.assert_called_with(2)
    assert result == 6

    out, err = capsys.readouterr()
    print(out)
    assert """\
<lambda>
started

<lambda>
>> transform
>> Load("dest")

etl
--------------------------------------------------------------------------------
<lambda>
>> transform
>> Load("dest")
""" == out


def test_workflow_as_flow(capsys):
    flow = Workflow() >> range(3) >> print >> range(6) >> print
    flow.run()

    out, err = capsys.readouterr()
    print(out)

    assert len(flow) == 4
    assert """\
range(0, 3)
>> print
range(0, 3)
>> range(0, 6)
>> print
range(0, 6)
""" == out


def test_workflow_with_flow(capsys):
    Workflow(
        Flow() >> range(3) >> print,
        another=Flow() >> range(6) >> print).run()

    out, err = capsys.readouterr()
    print(out)

    assert """\
range(0, 3)
>> print
range(0, 3)

another
--------------------------------------------------------------------------------
range(0, 6)
>> print
range(0, 6)
""" == out


def test_readme(capsys):
    def setup():
        print('Ready to go!')

    def extract():
        return 'data'

    def transform(data):
        return data + ' using reusable code pieces, like Lego.'

    class Load(Step):
        """ Sub-class Step to customize the callable """
        def __init__(self, to_dataset):
            super().__init__()
            self.to_dataset = to_dataset

        def __str__(self):
            return f'Load("{self.to_dataset}")'

        def run(self, data):
            print(f'Loading {data}')

    Workflow(setup,
             etl=[extract, transform, Load('example')]).run()

    print()
    flow = Workflow() >> extract >> transform >> Load('example')
    flow.run()

    out, err = capsys.readouterr()
    print(out)
    assert """\
setup
Ready to go!

etl
--------------------------------------------------------------------------------
extract
>> transform
>> Load("example")
Loading data using reusable code pieces, like Lego.

extract
>> transform
>> Load("example")
Loading data using reusable code pieces, like Lego.
""" == out



class Step:
    """ Abstract step that is callable, runnable, and iterable """
    def __init__(self, callable_or_obj=None):
        """
        :param callable|object callable_or_obj: A callable to be called or an object to be returned when step is run.
        """
        if isinstance(callable_or_obj, Step):
            raise ValueError(f'Must be a callable or an object, but got another Step instance: {callable_or_obj}')

        self.callable_or_obj = callable_or_obj

    def __call__(self, *args, **kwargs):
        self._print_name()
        return self.run(*args, **kwargs)

    def __iter__(self):
        return iter(self())

    @classmethod
    def instance(cls, obj):
        return obj if isinstance(obj, cls) else Step(obj)

    @property
    def name(self):
        """ Guess the name of the callable """
        if not hasattr(self, 'callable_or_obj'):
            raise RuntimeError('Step is not properly initialized. Please call `super().__init__()` in '
                               f'`{self.__class__.__name__}.__init__()`')
        co = self.callable_or_obj or self
        name = str(co)

        # Name given by the developer
        if 'at 0x' not in name and 'Mock name' not in name and 'built-in ' not in name and '<class ' not in name:
            return name

        # Name attributes
        if self.callable_or_obj:
            for name_attr in ('_mock_name', 'name', '__name__'):
                if hasattr(co, name_attr):
                    name = getattr(co, name_attr)
                    if name:  # Ensure it actually has a value
                        return name

        # Class name
        if hasattr(co, '__class__'):
            return co.__class__.__name__

        # Last resort
        return name

    def _print_name(self):
        print(self.name)

    def run(self, *args, **kwargs):
        if self.callable_or_obj:
            return self.callable_or_obj(*args, **kwargs) if callable(self.callable_or_obj) else self.callable_or_obj

        raise NotImplementedError('Please provide a callable or object when instantiating, or override this method.')


class Flow(list, Step):
    """ List of steps that will be run in serial with data being passed from one to another """
    def __init__(self, *steps, name=None):
        """
        :param list steps: List of steps for the flow
        :param str name: Name for the flow
        """
        super().__init__()
        if steps:
            self.extend([Step.instance(s) for s in steps])
        self._name = name

    def __rshift__(self, call):
        self.append(Step.instance(call))
        return self

    def _print_name(self):
        if self._name:
            print(self._name)
            print('-' * 80)

    @property
    def name(self):
        return self._name or len(self) and self[0].name or None

    def run(self):
        result = None

        for i, step in enumerate(self):
            if i == 0:
                result = step()

            else:
                print('>> ', end='')
                result = step(result)

        return result


class Workflow(Flow):
    """ Pythonic workflow engine that helps you write better ETL scripts  """
    def __init__(self, *steps, **flows):
        """
        :param list steps: List of steps (any callable) to run
        :param dict flows: Map of name to list of steps
        """
        super().__init__()
        self.steps = []

        for step in steps:
            if isinstance(step, (Step, Flow)):
                self.steps.append(step)
            elif callable(step):
                self.steps.append(Step(step))
            else:
                self.steps.append(Flow(*step))

        for name, flow in flows.items():
            self.steps.append(Flow(*flow, name=name))

    def run(self):
        """ Run all the steps """
        if self.steps:
            result = None

            for i, step in enumerate(self.steps):
                if i and isinstance(step, Flow):
                    print()
                result = step()

            return result

        self._print_name()
        return super().run()



import setuptools


setuptools.setup(
    name='datarunner',
    version='1.1.1',

    author='Max Zheng',
    author_email='maxzheng.os@gmail.com',

    description='A simple data workflow runner that helps you write better ETL scripts using reusable code pieces.',
    long_description=open('README.rst').read(),

    url='https://github.com/maxzheng/datarunner',

    install_requires=open('requirements.txt').read(),

    license='MIT',

    packages=setuptools.find_packages(),
    include_package_data=True,

    python_requires='>=3.6',
    setup_requires=['setuptools-git', 'wheel'],

    # entry_points={
    #    'console_scripts': [
    #        'script_name = package.module:entry_callable',
    #    ],
    # },

    # Standard classifiers at https://pypi.org/classifiers/
    classifiers=[
      'Development Status :: 5 - Production/Stable',

      'Intended Audience :: Developers',
      'Topic :: Software Development :: Libraries :: Python Modules',

      'License :: OSI Approved :: MIT License',

      'Programming Language :: Python :: 3',
      'Programming Language :: Python :: 3.6',
      'Programming Language :: Python :: 3.7',
    ],

    keywords='run call code workflow etl extract transform load',
)
